/**
 * Plain JavaScript library to allow adaptive sticky positioning.
 *
 * "Adaptive" means that an element should only actually be sticky if it passes
 * a check:
 * It has to fit into a maximal height (typically this is the viewport height).
 *
 * Concepts and caveats:
 * - we assume a big containing block (the flow root) with a direct (first)
 *   child, the sticky element
 * - the sticky element moves inside the containing block
 * - there is an optional target element used for offsetting the stickness --
 *   the target is *visually* the anchor of the stickyness (technically it is the
 *   sticky element)
 * - checking if really sticky (currently only maximum height)
 * - only px values supported yet (no need for other units yet)
 * - no transition on position (not easy to implement, and usally not needed)
 *
 * Usage example with jQuery:
 *
 *     const $containing = $('#grid-region > .inner')
 *     const $target = $('form', $containing)
 *
 *     $containing.on(
 *       'sticky:adaptiveSticky',
 *       (e) => {
 *         console.log(e)
 *       }
 *     )
 *
 *     sticky($containing[0], 'stick', {
 *       target: $target[0],
 *       maxHeight: window.innerHeight,
 *       debug: true
 *     })
 *
 * TODOs:
 * - add a class when initialized and/or active
 *
 * Maybe TODOs:
 *-  configurable checks
 * - accept rem, em,... too, not only px
 * - tests
 * - allow target and element to be querySelectors
 */

/**
 * Default values.
 *
 * - Expect integers for `offset` and `maxHeight`, which get interpreted as px
 *   values.
 * - `target` is expected to be an `HTMLElement`.
 */
const defaultOptions = {
  offset: 0,
  target: null,
  maxHeight: null,
  debug: false
}

/**
 * Storage for states of initialized stickies.
 */
const _state = new WeakMap()

/**
 * Parse a string for an integer, always return an integer.
 *
 * Converts `NaN` to 0.
 *
 * @param {string} string The string to parse.
 * @param {number} radix The radix.
 */
const parseIntOrZero = (string, radix = 10) => {
  const num = parseInt(string, radix)
  if (isNaN(num)) {
    return 0
  }
  return num
}

/**
 * Find the target given in the config.
 *
 * Currently this is just a sanitizing utility.
 *
 * @param {object} config The config object.
 */
const findTarget = config => {
  let target = null
  if (config.target) {
    if (typeof config.target === 'object' && config.target.nodeType === 1) {
      target = config.target
    }
  }
  return target
}

/**
 * Compute the offset for the sticky element.
 *
 * The offset is defined by a target (the element which should be the *visible*
 * sticky anchor), plus an offset to that anchor.
 *
 * If there is a `target` we have to calculate the target-relative offset
 * We use the viewport-relative data to achieve that:
 * https://developer.mozilla.org/en-US/docs/Web/API/Element/getBoundingClientRect
 *
 * @param {object} config The config object.
 * @param {HTMLElement} element The sticky element.
 * @param {HTMLElement|null} target An optional target for the offset.
 */
const computeOffset = (config, element, target) => {
  let top = 0
  if (target) {
    const viewportRelativePos1 = element.getBoundingClientRect()
    const viewportRelativePos2 = target.getBoundingClientRect()
    const diff = viewportRelativePos2.top - viewportRelativePos1.top
    top = diff
  }

  const offset = parseIntOrZero(config.offset)
  // Negative top because this is a *offset*
  const computedOffset = -(top + offset)
  return computedOffset
}

/**
 * Check if we actually want to got sticky.
 *
 * This is the "adaptive" part.
 *
 * We want to go sticky if the whole sticky area fits into the maximal height
 *
 * If no  maximal height given, go sticky by default.
 *
 * @param {HTMLElement} element The sticky element.
 * @param {number} computedOffset The computed offset as integer (px).
 * @param {number} maxHeight A maximal height to check.
 */
const checkGoSticky = (element, computedOffset, maxHeight) => {
  let goSticky = true
  if (typeof maxHeight === 'number') {
    const viewportRelativePos1 = element.getBoundingClientRect()
    // we can reuse computedOffset to respect the optional target
    const topPos = viewportRelativePos1.top - computedOffset
    const bottomPos = viewportRelativePos1.bottom
    const stickyHeight = bottomPos - topPos
    // check max height
    goSticky = maxHeight >= stickyHeight
  }
  return goSticky
}

/**
 *
 * @param {HTMLElement} element The sticky element.
 * @param {string} action The lifecycle action to perform on the sticky.
 * @param {object} options Configuration options.
 */
const sticky = (element, action, options = {}) => {
  const config = Object.assign({}, defaultOptions, options)
  // Check if element is an DOM Element
  if (typeof element !== 'object' || element.nodeType !== 1) {
    console.error('Given argument is not a DOM Element:', typeof element, element)
    return null
  }

  // A helper function for debug logging
  const logger = (...items) => {
    if (config.debug) {
      console.debug('[adaptiveSticky]', ...items)
    }
  }

  logger('init', action, element, config)

  /**
   * Return the internal state.
   *
   * This returns the full map of all initialized stickies.
   *
   * Helpful for debugging.
   */
  if (action === '_state') {
    return _state
  }

  /**
   * Check if the element has been initilized.
   */
  if (action === 'check') {
    return _state.has(element)
  }

  /**
   * Return the sticky status of the element.
   */
  if (action === 'status') {
    return _state.get(element)
  }

  /**
   * Initialize an element to be sticky.
   *
   * This has no effect if called multiple times on the same element.
   *
   * Use the action `update` to change elements behaviour.
   */
  if (action === 'stick') {
    if (!_state.has(element)) {
      const origPosition = element.style.position
      const origTop = element.style.top
      const s = {
        // the init has run
        sticky: true,
        // the checks passed, we actually are sticky
        active: null,
        // prior state, saved for reverting when unsticking
        origPosition: origPosition,
        origTop: origTop,
        // an Element to offset to if asked for
        targetElement: null,
        // computed offset of sticky element
        offset: 0,
        // current config, the default, init config and updated config combined
        config: config
      }

      const target = findTarget(config)
      const computedOffset = computeOffset(config, element, target)

      const goSticky = checkGoSticky(element, computedOffset, config.maxHeight)
      logger(action, 'goSticky', goSticky)

      if (goSticky) {
        // TODO: emit events
        element.style.position = 'sticky'
        element.style.top = `${computedOffset}px`
      }

      // trigger an event when the sticky state changes
      if (goSticky !== s.active) {
        const eventName = goSticky ? 'sticky' : 'unsticky'
        const event = new CustomEvent(`${eventName}:adaptiveSticky`, { bubbles: true })
        element.dispatchEvent(event)
      }

      // Update _state
      s.computedOffset = computedOffset
      s.active = goSticky
      s.targetElement = target
      _state.set(element, s)
    }
    return element
  }

  /**
   * Unstick an element.
   *
   * This reverts an element to it's former state, i.e. the state it had when
   * the action `stick` was called.
   *
   * Reverting covers the element style `position` and `top`.
   */
  if (action === 'unstick') {
    if (_state.has(element)) {
      element.style.position = _state.get(element).origPosition
      element.style.top = _state.get(element).origTop

      const eventName = 'unsticky'
      const event = new CustomEvent(`${eventName}:adaptiveSticky`, { bubbles: true })
      element.dispatchEvent(event)

      return _state.delete(element)
    }
    return false
  }

  /**
   * Update an sticky element.
   *
   * This does nothing if the element was not initialized with `stick`.
   *
   * Use this e.g. inside a `resize` listener, or with media queries.
   */
  if (action === 'update') {
    if (_state.has(element)) {
      const s = _state.get(element)
      const config = Object.assign({}, s.config, options)
      _state.set(element, s)

      const target = findTarget(config)
      const computedOffset = computeOffset(config, element, target)
      const goSticky = checkGoSticky(element, computedOffset, config.maxHeight)
      logger(action, 'goSticky', goSticky)

      if (goSticky) {
        // TODO: emit events
        element.style.position = 'sticky'
        element.style.top = `${computedOffset}px`
      } else {
        element.style.position = _state.get(element).origPosition
        element.style.top = _state.get(element).origTop
      }
      // trigger events when the sticky state changes
      if (goSticky !== s.active) {
        const eventName = goSticky ? 'sticky' : 'unsticky'
        const event = new CustomEvent(`${eventName}:adaptiveSticky`, { bubbles: true })
        element.dispatchEvent(event)
      }

      // Update _state
      s.computedOffset = computedOffset
      s.active = goSticky
      s.targetElement = target
      s.config = config
    }
    return element
  }

  // Wrong action given, return null
  return null
}

export { sticky }
